
/* 
 *   Use the lines following this comment block to set the type of EFIS (ie airspeed source) and LIDAR being used
 *   
 *   The following EFIS types are supported:
 *   
 *   EFIS_TYPE_ASX1   (MGL avionics combo altimeter and airspeed indicator
 *   EFIS_TYPE_SV     Dynon skyview ADHRS serial stream 
 *   EFIS_TYPE_G3X    Garmin G3X ADHRS serial stream 
 *   EFIS_TYPE_GRT    Grand Rapids Horizon/Sport outputting a Shadin FADC data stream 
 *   
 *   The following LIDARs are supported:
 *   LIDAR_TYPE_TF03        Bennewake TF03 
 *   LIDAR_TYPE_SF11        Lightware SF11 serial 
 *   LIDAR_TYPE_SF11_ANALOG Lightware SF11 configured to ouput an analog altitude signal  
 *   LIDAR_TYPE_SF30_ANALOG Lightware SF30 configured to output an analog altitude signal 
 *   
 *   
 */

 
#define EFIS_TYPE EFIS_TYPE_SV
#define LIDAR_TYPE LIDAR_TYPE_TF03

// uncomment the following line if the optional LCD display is (or will be) installed  
#define LCD_DISPLAY true

// specify which lcd display is being used 
#define LCD_DISPLAY_TYPE_DFROBOT true
//#define LCD_DISPLAY_TYPE_ADAFRUIT true

// uncomment the following line if the audio alert daughter card is installed 
#define AUDIO_ALERTS true 

// set the volume for the audio alerts - 

/**************************** WARNING **************************************************/
/******** DO NOT CHANGE ANYTHING BELOW HERE UNLESS YOU KNOW WHAT YOU ARE DOING *********/
/**************************** WARNING **************************************************/

// uncomment the following line if the serial monitor is attached
// (note: lidar serial input must be disconnected 
//#define SERIAL_MONITOR  true


/*
This sketch implements an automatic landing gear extension trigger device based on
airspeed, throttle setting, and LiDAR-determined AGL altitude. 

In particular, the trigger output line is set HIGH when the following conditions are raised:

 - airspeed lower than UPPER_AIRSPEED_THRESHOLD
 - airspeed higher than LOWER_AIRSPEED_THRESHOLD
 - altitude lower than UPPER_AGL_THRESHOLD
 - altitude  higher than LOWER_AGL_THRESHOLD
 - throttle LOW switch is ON

When all conditions are met the output trigger is set HIGH and remains latched until

Airspeed and RPM data are pulled from the EFIS datastream on Serial Stream 1
AGL altitude data is pulled from the Benewake LIDAR LIDAR unit on Serial Stream 2

All input data is is buffered using a 2-second moving average, sampled every 200 milliseconds

Input:

  Serial1_RX/Serial1_TX
  Serial2_RX/Serial2_TX
  AutoExEnable  (High when enabled, low when disabled)

Output:
  AutoExTrigger (High when all triggering conditions are met, remains latched until reset)
  LCD display   (optional remote 2x16 LCD display, shows altitude & airspeed and status/error messages 
  Audio Alerts  (audio callouts output to the stereo mini-jack)


Change Log:

  2022/12/12  Add message trailer functionality; improve efficiency on SerialWrapper update function
  2023/01/15  Add support for Audio callouts via DFRobot MP3 Player daughter card 
  2023/01/26  Add support for DFRobot 16x2 LCD i2c display module 
  
 */
 
#include <SoftwareSerial.h>

// lcd definitions (if installed)
#ifdef LCD_DISPLAY
#ifdef LCD_DISPLAY_TYPE_DFROBOT
#include <LiquidCrystal_I2C.h>
#endif
#ifdef LCD_DISPLAY_TYPE_ADAFRUIT
#include <Adafruit_LiquidCrystal.h>
#endif
#endif

// dfrobot mp3 player
#ifdef AUDIO_ALERTS 
#include <DFRobot_DF1201S.h>
#endif 
#include <Bounce2.h>
#include <limits.h>
#include <avr/pgmspace.h>



/* Pins */

/* Digital */
/* note pwm available on pins 3,5,6,9,10,11 */
/* digital inputs */
/* use hardware serial for efis */
#define LIDAR_RX 0
#define LIDAR_TX 1


/* software serial for lidar */
#define EFIS_RX 2
#define EFIS_TX 3

/* software serial for audio */

#define AEX_ENABLE_SW 4
#define GEAR_UP_SW_SENSE 5


/* digital outputs */
#define AEX_TRIGGER 6
#define THROTTLE_LOW_SW_SENSE 7

#define MP3_PLAYER_RX 8
#define MP3_PLAYER_TX 9


#define AEX_STATUS_LED 10

/* Analog */
#define  GEAR_UP_SNS A0
#define  GEAR_DN_SNS A1
#define  LIDAR_ANALOG_IN A2


#define  UNUSED_ANALOG_A3 A3

#define  LCD_DAT A4
#define  LCD_CLK A5


/* thresholds */


/* max LIDAR distance is 180m */

//#define UPPER_AIRSPEED_THRESHOLD 255
//#define LOWER_AIRSPEED_THRESHOLD 0
#define UPPER_AIRSPEED_THRESHOLD 90
#define LOWER_AIRSPEED_THRESHOLD 40

// set upper agl threshold to circuit altitude (none of the supported lidars gets close to this) 
#define UPPER_AGL_THRESHOLD 1000
#define LOWER_AGL_THRESHOLD 10

#define AIRSPEED_SENSOR_MAX 255
#define AIRSPEED_SENSOR_MIN 0
#define ALTITUDE_SENSOR_MAX 999
#define ALTITUDE_SENSOR_MIN 0


#define UPPER_RPM_THRESHOLD 0

// at least one working sensor is required  
#define AEX_MIN_SENSOR_CNT 1

#define VERSION_STR "Cozy Gear Auto-Extend  V1.0"
#define DEBOUNCE_INTERVAL 10

// smooth over last 10 readings 
#define SENSOR_BUFFER_CNT 10
// interval to determine if last sensor reading update was recent 
#define SENSOR_RECENT_MILLIS 1500

// keep track of 5 points for trend determination
#define TREND_POINT_CNT 5
// trend slopes within +/- of TREND_HYSTERESIS are considered CONSTANT  
#define TREND_HYSTERESIS 5
// trend enumerations 
#define TREND_INCREASING +1
#define TREND_CONSTANT 0
#define TREND_DECREASING -1
// record an altitude every second for trend detection 
#define ALT_TREND_CHECK_MILLIS 1000

  

/* switch modes  - normal mode a grounded switch (digital LOW) means swith is ON , inverted mode a digital HIGH = ON*/
#define INVERTED false
#define NORMAL   true

/* switch states */

#define ON true
#define OFF false

/* display characteristics  */
#define LCD_DISPLAY_COLS 16
#define LCD_DISPLAY_ROWS 2
#define LCD_LINE_1 0
#define LCD_LINE_2 1

/* AEX LED defs */
#define MAX_INTENSITY 255
#define AEX_STATUS_LED_INTENSITY 255
// no
#define LED_NOFLASH 0
// really slow flash (gear being deployed)
#define AEX_STATUS_LED_DEPLOY_FLASH_MILLIS 2000
// slow flash
#define AEX_STATUS_LED_INIT_FLASH_MILLIS 1000
// rapid flash
#define AEX_STATUS_LED_FAIL_FLASH_MILLIS 250
// warning flash
#define AEX_STATUS_LED_WARNING_FLASH_MILLIS 750
// super rapid flash
#define AEX_STATUS_LED_HARD_FAIL_FLASH_MILLIS 100

/* error codes */
#define OK 0
#define ENABLE true
#define DISABLE false
#define ON true
#define OFF false
#define ERROR -99
#define CONNECT_ERR -1
#define SERIAL_ERR -2
#define TIMEOUT_ERR -3

/* aex states */
#define AEX_OFF 0
#define AEX_ON 1
// recoverable failure (eg,sensors unresponsive)
#define AEX_FAIL -1
// armed and have both alt and spd sensors 
#define AEX_ARMED 2
// armed and have only one of alt/spd sensors 
#define AEX_ARMED_MIN_SENSOR 3

// unrecoverable failure, reboot required 
#define AEX_HARD_FAIL -2 

// check for state changes every tenth second
#define STATE_CHECK_MILLIS 100

// check for serial stream startup every second when in ON state
#define SERIAL_CHECK_MILLIS 1000

// stream timeout millis  - should get a new message at least every  couple of seconds 
#define STREAM_TIMEOUT_MILLIS  2000
/*
    Serial stream name equates for different types of lidar and efis datastreams 
*/

#define EFIS_TYPE_ASX1 0
#define EFIS_TYPE_SV   1
#define EFIS_TYPE_G3X  2
// note grt serial port should be configured to ouput Shadin type Z data stream
// this is also compatible with a Shadin fuel/air or airdata unit and a number of other devices 
#define EFIS_TYPE_GRT  3

// benewake TF03 (180m range)
#define LIDAR_TYPE_TF03 0
// lightware SF11 (100m range)
#define LIDAR_TYPE_SF11 1
// lightware SF11 (100m range) using the analog output instead of serial data stream
#define LIDAR_TYPE_SF11_ANALOG 2 
// terra ranger evo  (60m range) 
#define LIDAR_TYPE_SF30_ANALOG 3 
// terra ranger evo  (60m range) 
#define LIDAR_TYPE_TREV 4



//#define LIDAR_STREAM Serial
#define MAX_MSG_LEN 80
//#define LIDAR_STREAM lidarSerial
#define LIDAR_STREAM Serial
#define EFIS_STREAM efisSerial

// stream timeout is two seconds
#define STREAM_TIMEOUT_MILLIS 2000

// TF03 equates
#define TF03_ALT_MSB_IDX  3
#define TF03_ALT_LSB_IDX  2
#define TF03_QOM_MSB_IDX  5
#define TF03_QOM_LSB_IDX  4
#define TF03_ALT_MAX 585
#define TF03_ALT_MIN 0
#define TF03_MSG_HDR '\x59','\x59'
#define TF03_MSG_LEN 9
#define TF03_MSG_CHECKSUM_START_IDX 0 
#define TF03_MSG_CHECKSUM_STOP_IDX 0
#define TF03_MSG_CHECKSUM_IDX 8
#define TF03_FT_CONVERSION_FACTOR .0328
#define TF03_BAUDRATE 115200


// SF11 lidar equates
#define SF11_ALT_MSB_IDX  2
#define SF11_ALT_LSB_IDX  7
// max alt for SF11 is 100m (328 ft). unit returns 130M when it is out of range 
#define SF11_ALT_MAX 328 
#define SF11_ALT_MIN 0
// SF11 just sends a continous stream of ascii numbers, but each on ends in a cr/lf
// so we treat it as a header for message separation purposes 
#define SF11_MSG_HDR '\x0d','\x0a'
// SF11 max numeric value is 130.00, so 8 is max msg len when header is added 
#define SF11_MSG_LEN 8
// no checksum for SF11 in streaming mode
#define SF11_MSG_CHECKSUM_START_IDX 0 
#define SF11_MSG_CHECKSUM_STOP_IDX 0
#define SF11_MSG_CHECKSUM_IDX 0
#define SF11_FT_CONVERSION_FACTOR .0328
#define SF11_BAUDRATE 19200

// SF11 analog lidar equates same as for SF11 except for conversion factor. SF11 analog values go from 0 to 3.3 volts, 
// corresponding to 10-bit adc value of 675. 100M = 328ft so multiply or adc value by this amount to get corresponding alt in ft.  

#define SF11_ANALOG_FT_CONVERSION_FACTOR .486

// SF30 analog lidar equates same as for SF11 except for conversion factor (10-bit adc value of 675 = 200m =  656 ft)
#define SF30_ANALOG_FT_CONVERSION_FACTOR .972


// ASX-1 equates
#define ASX1_SPD_MSB_IDX  6
#define ASX1_SPD_LSB_IDX  7
#define ASX1_SPD_MIN 0
#define ASX1_SPD_MAX 255
#define ASX1_MSG_HDR  '\x82','\xFF'
#define ASX1_MSG_LEN 11
#define ASX1_MSG_CHECKSUM_START_IDX 0 
#define ASX1_MSG_CHECKSUM_LEN  0
#define ASX1_MSG_CHECKSUM_IDX 9
#define ASX1_KT_CONVERSION_FACTOR .87
#define ASX1_BAUDRATE 19200

// skyview adahrs data stream equates 
#define SV_SPD_MSB_IDX  23
#define SV_SPD_LSB_IDX  26
#define SV_SPD_MIN 0
#define SV_SPD_MAX 255
#define SV_MSG_HDR  '!','1'
#define SV_MSG_LEN 74
#define SV_MSG_CHECKSUM_START_IDX  0
#define SV_MSG_CHECKSUM_LEN  70
#define SV_MSG_CHECKSUM_IDX 70
#define SV_KT_CONVERSION_FACTOR .1
#define SV_BAUDRATE 9600

// g3x adahrs data stream equates 
#define G3X_SPD_MSB_IDX  23
#define G3X_SPD_LSB_IDX  26
#define G3X_SPD_MIN 0
#define G3X_SPD_MAX 255
#define G3X_MSG_HDR  '=','1'
#define G3X_MSG_LEN 59
#define G3X_MSG_CHECKSUM_START_IDX  0
#define G3X_MSG_CHECKSUM_LEN  55
#define G3X_MSG_CHECKSUM_IDX 55
#define G3X_KT_CONVERSION_FACTOR .1
#define G3X_BAUDRATE 19200

// grt adahrs data stream shadin 'Z'  equates 
#define GRT_SPD_MSB_IDX  3
#define GRT_SPD_LSB_IDX  5
#define GRT_SPD_MIN 0
#define GRT_SPD_MAX 255
#define GRT_MSG_HDR  'x\02'
#define GRT_MSG_TRLR 'x\03'
// variable message length (potentially)
#define GRT_MSG_LEN 0
#define GRT_MSG_CHECKSUM_START_IDX  0
#define GRT_MSG_CHECKSUM_LEN  3
#define GRT_MSG_CHECKSUM_IDX 0
#define GRT_KT_CONVERSION_FACTOR 1
#define GRT_BAUDRATE 9600

// start of TF03 LIDAR equates 
#if LIDAR_TYPE == LIDAR_TYPE_TF03
#define ALT_MSB_IDX TF03_ALT_MSB_IDX
#define ALT_LSB_IDX TF03_ALT_LSB_IDX
#define ALT_MAX TF03_ALT_MAX
#define ALT_MIN TF03_ALT_MIN
#define ALT_MSG_HDR TF03_MSG_HDR
#define ALT_MSG_LEN TF03_MSG_LEN
#define ALT_MSG_CHECKSUM_START_IDX TF03_MSG_CHECKSUM_START_IDX  
#define ALT_MSG_CHECKSUM_LEN TF03_MSG_CHECKSUM_LEN  
#define ALT_MSG_CHECKSUM_IDX TF03_MSG_CHECKSUM_IDX 
#define ALT_FT_CONVERSION_FACTOR TF03_FT_CONVERSION_FACTOR
#define LIDAR_STREAM_BAUDRATE TF03_BAUDRATE
#endif
// end of TF03 LIDAR equates 

// start of SF11/SF30 LIDAR equates 
#if (LIDAR_TYPE == LIDAR_TYPE_SF11) || (LIDAR_TYPE == LIDAR_TYPE_SF11_ANALOG) || (LIDAR_TYPE == LIDAR_TYPE_SF30_ANALOG)
#define ALT_MSB_IDX SF11_ALT_MSB_IDX
#define ALT_LSB_IDX SF11_ALT_LSB_IDX
#define ALT_MAX SF11_ALT_MAX
#define ALT_MIN SF11_ALT_MIN
#define ALT_MSG_HDR SF11_MSG_HDR
#define ALT_MSG_LEN SF11_MSG_LEN
#define ALT_MSG_CHECKSUM_START_IDX SF11_MSG_CHECKSUM_START_IDX  
#define ALT_MSG_CHECKSUM_LEN SF11_MSG_CHECKSUM_LEN  
#define ALT_MSG_CHECKSUM_IDX SF11_MSG_CHECKSUM_IDX 
#define ALT_FT_CONVERSION_FACTOR SF11_FT_CONVERSION_FACTOR
#define LIDAR_STREAM_BAUDRATE SF11_BAUDRATE
#endif
#if (LIDAR_TYPE == LIDAR_TYPE_SF11_ANALOG)
#define ALT_FT_CONVERSION_FACTOR SF11_ANALOG_FT_CONVERSION_FACTOR
#endif
#if (LIDAR_TYPE == LIDAR_TYPE_SF30_ANALOG)
#define ALT_FT_CONVERSION_FACTOR SF30_ANALOG_FT_CONVERSION_FACTOR
#endif
// end of SF11 LIDAR equates


// start of ASX1 equates  
#if EFIS_TYPE == EFIS_TYPE_ASX1
#define SPD_MSB_IDX ASX1_SPD_MSB_IDX  
#define SPD_LSB_IDX ASX1_SPD_LSB_IDX  
#define SPD_MIN ASX1_SPD_MIN 
#define SPD_MAX ASX1_SPD_MAX
#define SPD_MSG_HDR ASX1_MSG_HDR  
#define SPD_MSG_LEN ASX1_MSG_LEN
#define SPD_MSG_CHECKSUM_START_IDX ASX1_MSG_CHECKSUM_START_IDX 
#define SPD_MSG_CHECKSUM_LEN ASX1_MSG_CHECKSUM_LEN 
#define SPD_MSG_CHECKSUM_IDX ASX1_MSG_CHECKSUM_IDX
#define SPD_KT_CONVERSION_FACTOR ASX1_KT_CONVERSION_FACTOR 
#define EFIS_STREAM_BAUDRATE 19200
#endif
// end of ASX1 equates 

// start of SV equates  
#if EFIS_TYPE == EFIS_TYPE_SV
#define SPD_MSB_IDX SV_SPD_MSB_IDX  
#define SPD_LSB_IDX SV_SPD_LSB_IDX  
#define SPD_MIN SV_SPD_MIN 
#define SPD_MAX SV_SPD_MAX
#define SPD_MSG_HDR SV_MSG_HDR  
#define SPD_MSG_LEN SV_MSG_LEN
#define SPD_MSG_CHECKSUM_START_IDX SV_MSG_CHECKSUM_START_IDX 
#define SPD_MSG_CHECKSUM_LEN SV_MSG_CHECKSUM_LEN 
#define SPD_MSG_CHECKSUM_IDX SV_MSG_CHECKSUM_IDX
#define SPD_KT_CONVERSION_FACTOR SV_KT_CONVERSION_FACTOR 
#define EFIS_STREAM_BAUDRATE SV_BAUDRATE
#endif
// end of SV equates 

// start of G3X equates  
#if EFIS_TYPE == EFIS_TYPE_G3X
#define SPD_MSB_IDX G3X_SPD_MSB_IDX  
#define SPD_LSB_IDX G3X_SPD_LSB_IDX  
#define SPD_MIN G3X_SPD_MIN 
#define SPD_MAX G3X_SPD_MAX
#define SPD_MSG_HDR G3X_MSG_HDR  
#define SPD_MSG_LEN G3X_MSG_LEN
#define SPD_MSG_CHECKSUM_START_IDX G3X_MSG_CHECKSUM_START_IDX 
#define SPD_MSG_CHECKSUM_LEN G3X_MSG_CHECKSUM_LEN 
#define SPD_MSG_CHECKSUM_IDX G3X_MSG_CHECKSUM_IDX
#define SPD_KT_CONVERSION_FACTOR G3X_KT_CONVERSION_FACTOR 
#define EFIS_STREAM_BAUDRATE G3X_BAUDRATE
#endif
// end of G3X equates

// start of GRT equates  
#if EFIS_TYPE == EFIS_TYPE_GRT
#define SPD_MSB_IDX GRT_SPD_MSB_IDX  
#define SPD_LSB_IDX GRT_SPD_LSB_IDX  
#define SPD_MIN GRT_SPD_MIN 
#define SPD_MAX GRT_SPD_MAX
#define SPD_MSG_HDR GRT_MSG_HDR  
#define SPD_MSG_TRLR GRT_MSG_TRLR
#define SPD_MSG_LEN GRT_MSG_LEN
#define SPD_MSG_CHECKSUM_START_IDX GRT_MSG_CHECKSUM_START_IDX 
#define SPD_MSG_CHECKSUM_LEN GRT_MSG_CHECKSUM_LEN 
#define SPD_MSG_CHECKSUM_IDX GRT_MSG_CHECKSUM_IDX
#define SPD_KT_CONVERSION_FACTOR GRT_KT_CONVERSION_FACTOR 
#define EFIS_STREAM_BAUDRATE GRT_BAUDRATE
// increase efis message size buffer to accomodate GRT messages 
#define MAX_MSG_LEN 140 
#endif
// end of G3X equates

// status message field equates  
#define STATUS_MSG_Q_LEN 3
#define STATUS_MSG_DISPLAY_MILLIS 3000



class Led
{
    // Class Member Variables
    // These are initialized at startup
    int ledPin;      // the number of the LED pin, should be a PWM pin

    int Intensity;  // duty cycle
    int ledState;                  // ledState used to set the LED
    int  flashMillis;     // flash rate in milliseconds
    unsigned long flashTimer;      // future time to flip state (when flashing)
    
    // Constructor - creates a Flasher
    // and initializes the member variables and state
public:
    Led(int pin)
    {
        ledPin = pin;
        pinMode(ledPin, OUTPUT);
        
        flashMillis = 0;
        
        ledState = LOW;
        flashTimer = 0;
        
        Intensity = MAX_INTENSITY  ;   // default to max
        digitalWrite(ledPin, LOW);     // turn off LED
    }
    // start flashing the led at interval inMillis (do nothing if the flash interval is same as current)
    void setFlash(int inMillis)
    {
        if (inMillis != flashMillis)
        {
          ledState = HIGH;       // set Led state to known value
          flashMillis = inMillis;
          flashTimer = millis() + flashMillis;
        }
    }
    void setIntensity(int intensity)
    {
        Intensity = intensity;

    }
    void turnOn()
    {
        flashMillis = 0;
        ledState = HIGH;
    }
    void turnOff()
    {
        flashMillis = 0;
        ledState = LOW;

    }
    void update()
    {

        // if we are flashing the LED, check to see if it's time to change the state of the LED
        if (flashMillis > 0 && millis() >= flashTimer)
        {
          ledState = (ledState == HIGH) ? LOW : HIGH;
          flashTimer = millis() + flashMillis;  // reset the flashtimer
        }
        if (ledState == HIGH)
        {
          if (Intensity == MAX_INTENSITY )
          {
            digitalWrite(ledPin,HIGH);  // don't use PWM if at max intensity
          }
          else
          {
            analogWrite(ledPin, Intensity);  // turn on LED at current intensity level
          }
        }
        else
        {
          if (Intensity == MAX_INTENSITY )
          {
            digitalWrite(ledPin,LOW);  // don't use PWM if at max intensity
          }
          else
            analogWrite(ledPin, 0);  // turn off PWM LED
          }

        }
    
};
// sensor values are positive by definition, so use -1 as unknown/unavailable value
#define UNKNOWN -1   
/* smoothed digital sensor */
class BufferedSensor {
  private:
    int sensorValues[SENSOR_BUFFER_CNT];
    long unsigned int sensorTimes[SENSOR_BUFFER_CNT]; 
    int valuesNext = 0;    // next slot to use
    int valuesCount = 0;   // count of slots used
    int pin;                   // pin assigned to this sensor
    int lastValue = 0;
    int bufferedValue = 0;
    int lowLimit = 0;    // reject any values below this
    int highLimit = 1023;   // reject any values above this
    long unsigned int updateMillis = 0;  // time since last update (used to determine if data is stale)
    bool enableState;           // enable/disable     
  public:
    BufferedSensor(int initLowLimit, int initHighLimit) 
    {
      /* init the ring buffer upon instantiation */
      int i;
      for (i = 0; i < SENSOR_BUFFER_CNT; i++)
      {
        sensorValues[i] = 0;
      }
      lowLimit = initLowLimit;
      highLimit = initHighLimit;
      enableState = false;        // disabled by default
    };

    // insert a value into the ring buffer and recalculate the rolling average 
    // note: a disabled sensor can still be updated 
    void update(int nextValue)
    {
      int totalValue = 0;
      int i;
      
      updateMillis = millis();
      lastValue = nextValue;  // default to using input value 
      //set values which are avove or below the sensor limit to the sensor limit
      lastValue = (nextValue < lowLimit)  ? lowLimit : nextValue; 
      lastValue = (nextValue > highLimit) ? highLimit : nextValue; 
      
      sensorValues[valuesNext] = lastValue;
      sensorTimes[valuesNext] = updateMillis; 
      valuesNext++;
      valuesNext %= SENSOR_BUFFER_CNT;
      valuesCount = (++valuesCount >= SENSOR_BUFFER_CNT) ? SENSOR_BUFFER_CNT : valuesCount; // this is just for when we are initially filling the buffer
      for (i = 0; i < valuesCount; i++)
      {
        totalValue = totalValue + sensorValues[i];
      }
      bufferedValue = int(totalValue / valuesCount);
      

    }
    int getInstantValue() {
      int myValue;
      myValue = (enableState) ? lastValue : UNKNOWN;
      return myValue;
    };
    int getValue()
    {
      int myValue;
      myValue = (enableState) ? bufferedValue : UNKNOWN;
      return myValue;
    }
    void setLimits(int newLowLimit , int newHighLimit)
    {
      lowLimit = newLowLimit;
      highLimit = newHighLimit;
    }
    long unsigned int getLastUpdateTime()
    {
      return updateMillis;
    }
    bool isRecent()
    {
      bool returnVal = false;
      //Serial.println(millis() - updateMillis);
      if (((millis() - updateMillis) < SENSOR_RECENT_MILLIS) && enableState)
      {
        returnVal = true;
      }
      return returnVal;
    }
    bool getEnableState()
    {
      return (enableState);
    }
    
    void enable()
    {
      enableState = true;  
    }
    
    void disable()
    {
      enableState = false;  
    }  

    /* determine if the sensor trend is increasing, stable, or decreasing using least squares regression */

   
};


/*
  wrapper class so we can use it for both hard and soft
  serial
*/
 

class serialWrapper
{
  private:
  HardwareSerial* hwStream = NULL;
  SoftwareSerial* swStream = NULL;
  Stream* port = NULL;
  byte msgHdr[4] = {'\x00', '\x00','\x00','\x00'};
  int msgHdrLen = 0;
  byte msgTrlr[4] = {'\x00', '\x00','\x00','\x00'};
  int msgTrlrLen = 0;
  byte rcvBuf[MAX_MSG_LEN];  // read stream into this buffer
  int rcvBufPos = 0; // current cursor
  byte msgBuf[MAX_MSG_LEN];  // holds completed message
  byte inByte;
  unsigned int msgLen = 0;
  bool newMsg = false;  // indicates if message has been seen or not
  byte hexBuf[3];
  int bufOverrunCnt ; 
  int msgSkipCnt;
  int rcvBufState;
  long unsigned int msgMillis;    // time last message was received 
  #define RCV_BUF_NO_MSG 0
  #define RCV_BUF_IN_MSG 1
  public:
  serialWrapper (HardwareSerial& device) {hwStream = &device;}
  serialWrapper (SoftwareSerial& device) {swStream = &device;}

	void begin(unsigned long int baudRate)
	{
		if (hwStream)
		{
		  hwStream->begin(baudRate);
		  port = hwStream;
		}
		else
		{
		  //Serial.println("starting sw serial at baudrate"); Serial.println(baudRate);
		  swStream->begin(baudRate);
		  port = swStream;
		}
		rcvBufPos = bufOverrunCnt = msgSkipCnt = 0;
		rcvBufState = RCV_BUF_NO_MSG;
	}

	void print (String msg)
	{
		port->print(msg);
	}
	void print (int num)
	{
		port->print(num);
	}
	void println (String msg)
	{
		port->println(msg);
	}
	void end()
  	{     
    	if (hwStream) {hwStream->end();}
    	if (swStream) {swStream->end();}
    	port = NULL;
  	}
  
  	bool status()
  	{
    	bool portStatus = false;
    	if (port != NULL) portStatus = true;
    	return portStatus;
  	}
  
  	void setMsgHdr(char *hdr, int hdrLen)
  	{

    	msgHdrLen = hdrLen;
    	memcpy(msgHdr,hdr,hdrLen);
    	//Serial.println("msg header is: "); Serial.print(msgHdr); Serial.print(hdr);
    	return;
	}
  	void setMsgTrlr(char *trlr, int trlrLen)
  	{

    	msgTrlrLen = trlrLen;
    	memcpy(msgTrlr,trlr,trlrLen);
    	//Serial.println("msg header is: "); Serial.print(msgHdr); Serial.print(hdr);
    	return;
	}
	byte read()	
	{
    	if (port)
    	{
      	inByte = port->read();
    	}
    	return inByte;
	}
	int available()
  {   
    	if (port)
    	{
      	return(port->available());
    	}
	}
	void update()
	{
	
 		if (port)   // port is open
		{
			while (port->available() > 0)  // we have at least 1 byte available		
			{
        
				 //Serial.print(port->read(),HEX);
				 //Serial.println();
				if (rcvBufPos < MAX_MSG_LEN)
				{
					 rcvBuf[rcvBufPos++] = port->read();
					 
				   //Serial.print("byte received: "); Serial.println(rcvBuf[rcvBufPos], HEX);
					 // if most recent byte read equals the last byte of the header, see if we have run into a new header string
					 if ((rcvBuf[rcvBufPos-1] == msgHdr[msgHdrLen - 1]) && (rcvBufPos >= msgHdrLen))  
					 {
						if (strncmp(&rcvBuf[rcvBufPos - msgHdrLen],msgHdr,msgHdrLen) == 0)
						{
              //Serial.print("header found at position: "); Serial.println(rcvBufPos - msgHdrLen);
							// header found
							switch (rcvBufState)
							{
								// not in a message - start a new one by discarding anything already in the receive buffer and copy header to start of rcvbuf 
								case RCV_BUF_NO_MSG:
                  //Serial.print("rcvbufnomsg, rcvbufpos is:"); Serial.println(rcvBufPos);
									rcvBufState = RCV_BUF_IN_MSG;
									memcpy(rcvBuf,&rcvBuf[rcvBufPos-msgHdrLen],msgHdrLen);
                  //Serial.println((char *)rcvBuf);
									rcvBufPos = msgHdrLen;
									break;
								// already in a message, copy current message to msgBuf 	
								case RCV_BUF_IN_MSG:
                  //Serial.print("rcvbufinmsg, rcvbufpos is:"); Serial.println(rcvBufPos);
									if (msgTrlrLen == 0)    // no trailer defined , messages delineated by hdr only. Start a new message and copy current message to msg buf 
									{
										msgLen = rcvBufPos - msgHdrLen;	
                    //Serial.print("msglen is:"); Serial.println(msgLen);

										memcpy(msgBuf,rcvBuf,msgLen);
                    //Serial.println((char *)msgBuf);
										if (newMsg) msgSkipCnt++;   // increment skipcnt if msg in msgbuf has not been retrieved yet 
										newMsg = true; 				// flag msg available for pickup
                    msgMillis = millis();       // record time of last message 
										rcvBufPos = msgHdrLen;      // point past hdr
                    //for(int i = 0; i < msgLen; i++) {Serial.println(msgBuf[i],HEX);}
									}	
									break;												
							} // end switch 
						} // eif header string match 
					} // eif current byte matches end of hdr string 
			
					// message trailer defined and we just encountered a byte matching last byte of it 	
					if ((msgTrlrLen > 0) && (rcvBuf[rcvBufPos-1] == msgTrlr[msgTrlrLen-1]) && (rcvBufPos >= msgTrlrLen))   		
					{
						if ((strncmp(&rcvBuf[rcvBufPos - msgTrlrLen],msgTrlr,msgTrlrLen) == 0) && (rcvBufState == RCV_BUF_IN_MSG))
						{

							msgLen = rcvBufPos;
              //Serial.print("trailer found, msglen is:"); Serial.println(msgLen);	

							memcpy(msgBuf,rcvBuf,msgLen);
              //Serial.println((char *)msgBuf);
							if (newMsg) msgSkipCnt++;   // increment skipcnt if msg in msgbuf has not been retrieved yet 
							newMsg = true; 				// flag msg available for pickup
              msgMillis = millis();   // record time of new message 
							rcvBufPos = 0;
							rcvBufState = RCV_BUF_NO_MSG;
						} // eif trailer string match 
					} // eif current byte matches end of trailer string 
			 
				} // eif rcvBufPos < MAX_MSG_LEN
		
				// next read will overrun the buffer - reset receive buffer state, discard current message if one in progress 
				else
				{	
          //Serial.println("Buffer overflow detected");
					bufOverrunCnt++;
					if (rcvBufState == RCV_BUF_IN_MSG) msgSkipCnt++;
					rcvBufState = RCV_BUF_NO_MSG;
					rcvBufPos = 0;			
				} 			 
			} // ewhile byte available 
		}  // eif port is open     
		return;
	}

	int  getNewMsg(byte* outMsgBuf)
  {
    
	   memcpy(outMsgBuf,msgBuf,msgLen);
     newMsg = false;
     return msgLen;

    }

    bool haveNewMsg()
    {
      return newMsg;
    }

    long unsigned int getMsgMillis()
    {
      return msgMillis;
    }
}
;


/*
  non-precision timer (poll based, not interrupt)
  good for led flashers and the like
 */
 
class Timer {
  private:
    unsigned long int popMillis;
    unsigned int intervalMillis;
    bool timerPop;
    bool timerRunning;
  public:
    Timer() {

    };

    void update()
    {
      if (timerRunning && millis() > popMillis)
      {
         timerPop = true;
      }
    };
    void start(unsigned long timerInterval)
    {
      intervalMillis = timerInterval;
      popMillis = millis() + intervalMillis;
      timerRunning = true;
      timerPop = false;
    };
    void stop()
    {
      popMillis = millis() ;
      timerRunning = false;
      timerPop = false;
    };
    bool isExpired()
    {
      return timerPop;
      
    }
    bool isRunning()
    {
      return timerRunning;
    }

    
};

class Trend {
  private:
    int trendValues[TREND_POINT_CNT]; 
    int nextPoint = 0; 
    bool trendValid =false ; 
    
  public:
    Trend() {
      for (int i = 0; i < TREND_POINT_CNT; i++)
      {
        trendValues[i] = -1;  // special initial value 
      }
      nextPoint = 0;
      trendValid = false;
    };

    void update(int newValue)
    {
      trendValues[nextPoint] = newValue;
      nextPoint = (nextPoint+1) % TREND_POINT_CNT; 
      if (nextPoint == 0) trendValid = true;   // we have wrapped the trend buffer at least once 
    };

    bool getTrendState()
    {
      return(trendValid);
    }  
    
    int getTrend()
    {
      int trendValue, valueDelta, newestValueIdx;
      // nextPoint has oldest value, nextPoint - 1 has newest 
      newestValueIdx = ((nextPoint - 1) < 0) ? (TREND_POINT_CNT - 1) : (nextPoint - 1);
      valueDelta = trendValues[nextPoint] - trendValues[newestValueIdx]  ;
      //Serial.print("valueDelta: "); Serial.println(valueDelta);
      if ((valueDelta < 0) && (-valueDelta > TREND_HYSTERESIS)) trendValue = TREND_INCREASING ;   
      else if ((valueDelta > 0) && (valueDelta > TREND_HYSTERESIS)) trendValue = TREND_DECREASING ;
      else trendValue = TREND_CONSTANT; 
      
      return trendValue;    
    }

    
};
// global variables
int gCurrentAirspeed = UNKNOWN;
int gCurrentAltitude = UNKNOWN;
bool gAexTrigger= OFF;
int gAexState = AEX_OFF;




static byte gsAltHdr[] = {ALT_MSG_HDR};
static byte gsSpdHdr[] = {SPD_MSG_HDR};
#ifdef SPD_MSG_TRLR
static byte gsSpdTrlr[] = {SPD_MSG_TRLR};
#endif 

struct statusMsg {
     char msg[16];
     unsigned long createTime;
    bool displayed;
};

#ifdef LCD_DISPLAY
#ifdef LCD_DISPLAY_TYPE_ADAFRUIT
Adafruit_LiquidCrystal lcd(0);
#endif
#ifdef LCD_DISPLAY_TYPE_DFROBOT
LiquidCrystal_I2C lcd(0x20,16,2);
#endif

statusMsg gStatusMsgQueue[STATUS_MSG_Q_LEN]; 
#endif


/*  global variables and structures  */
SoftwareSerial EFIS_STREAM(EFIS_RX,EFIS_TX);
serialWrapper efis(EFIS_STREAM);
//SoftwareSerial LIDAR_STREAM(LIDAR_RX,LIDAR_TX);
serialWrapper lidar(LIDAR_STREAM);

#ifdef AUDIO_ALERTS 
SoftwareSerial mp3PlayerSerial(MP3_PLAYER_RX, MP3_PLAYER_TX);
DFRobot_DF1201S mp3Player;
bool gHaveMp3Player = false;
#endif 

BufferedSensor spd = BufferedSensor(AIRSPEED_SENSOR_MIN,AIRSPEED_SENSOR_MAX);
BufferedSensor alt = BufferedSensor(ALTITUDE_SENSOR_MIN,ALTITUDE_SENSOR_MAX);

Trend altTrend = Trend();

Bounce aexEnableSw ;
Bounce gearUpSw ;
Bounce throttleLowSw;

Led aexStatusLed  = Led(AEX_STATUS_LED);

Timer statusMsgTimer = Timer();
Timer stateCheckTimer = Timer();
Timer serialStartCheckTimer = Timer();
Timer altTrendTimer = Timer();

long unsigned int gAltCheckMillis;
long unsigned int gGearCheckMillis; 

// info messges (stored in flash) 
char  gFlashBuf[] =                 "1234567890123456";    // messages are read into here from flash
const PROGMEM char fEnableMsg[] =     "AEX Enabled     ";
const PROGMEM char fDisableMsg[] =    "AEX Disabled    ";
const PROGMEM char fFailMsg[] =       "AEX Failed      ";
const PROGMEM char fTriggerMsg[] =    "AEX Triggered   ";
const PROGMEM char fArmMsg[] =        "AEX Armed       ";
const PROGMEM char fArmWarnMsg[] =    "AEX Armed - WARN";
const PROGMEM char fExtendMsg[] =     "Extending Gear  ";
const PROGMEM char fExtCancelMsg[] =  "Gear ext.cancled";
const PROGMEM char fEfisToMsg[] =     "spd data timeout";
const PROGMEM char fLidarToMsg[] =    "alt data timeout";
const PROGMEM char fWaitSensorMsg[] = "Wt: spd/alt data";
const PROGMEM char fArmNoSpdMsg[] =   "WRN:Armed no spd";
const PROGMEM char fArmNoAltMsg[] =   "WRN:Armed no alt";
const PROGMEM char fErrNoSensorMsg[] ="Er:no spd,no alt";
const PROGMEM char fAudioFailMsg[] =  "Er:audioini fail";

// audio file index 
char gAudioFileNameBuf = "/xxx.MP3";
const PROGMEM char fOneFoot[] =       "/001.mp3";     // one 
bool gOneFootPlayed = false;
const PROGMEM char fTwoFoot[] =       "/002.mp3";     // two 
bool gTwoFootPlayed = false;
const PROGMEM char fThreeFoot[] =     "/003.mp3";     // three
bool gThreeFootPlayed = false;
const PROGMEM char fFourFoot[] =      "/004.mp3";     // four 
bool gFourFootPlayed = false;
const PROGMEM char fFiveFoot[] =      "/005.mp3";     // five feet
bool gFiveFootPlayed = false;
const PROGMEM char fTenFoot[] =       "/010.mp3";     // ten feet
bool gTenFootPlayed = false;
const PROGMEM char fTwentyFoot[] =    "/020.mp3";     // twenty feet
bool gTwentyFootPlayed = false;
const PROGMEM char fFiftyFoot[] =     "/050.mp3";     // fifty feet
bool gFiftyFootPlayed = false;
const PROGMEM char fOneHundredFoot[] = "/100.mp3";    // one hundred feet
bool gOneHundredFootPlayed = false;
const PROGMEM char fTwoHundredFoot[] = "/200.mp3";    // two hundred feet
bool gTwoHundredFootPlayed = false;
const PROGMEM char fGearUp[] =        "/201.mp3";     // gear is up
const PROGMEM char fGearDn[] =        "/202.mp3";     // gear is down
const PROGMEM char fAexEnable[] =     "/203.mp3";     // auto-extend is enabled 
const PROGMEM char fAexDisable[] =    "/204.mp3";     // auto-extend disabled 
const PROGMEM char fGearExtending[] = "/205.mp3";     // gear is extending
const PROGMEM char fGearRetracting[] ="/206.mp3";     // gear is retracting 
const PROGMEM char fGearUpWarn[] =    "/207.mp3";     // warning: check nose gear is down
const PROGMEM char fAexArm[] =       "/208.mp3";      // auto-extend armed    
const PROGMEM char fGearDownWarn[] = "/209.mp3";      // Reminder: retract nose gear 
bool gGearDownWarnPlayed = false;
const PROGMEM char fAudioEnabled[] = "/210.mp3";      // audio alerts enabled (startup)
#define AUDIO_FILE_NAME_CNT 8

void setup()
{


  gAexState = AEX_OFF;
#ifdef SERIAL_MONITOR  
  Serial.begin(115200);
  Serial.println("Hi there");     
#endif

  // set up pinmodes
  pinMode(LIDAR_RX,INPUT);
  pinMode(LIDAR_TX,OUTPUT);

  pinMode(EFIS_RX,INPUT);
  pinMode(EFIS_TX,OUTPUT);

  pinMode(AEX_ENABLE_SW,INPUT);
  pinMode(GEAR_UP_SW_SENSE,INPUT_PULLUP);
  pinMode(THROTTLE_LOW_SW_SENSE,INPUT_PULLUP);

  aexEnableSw.attach(AEX_ENABLE_SW,INPUT);
  aexEnableSw.interval(DEBOUNCE_INTERVAL);
  gearUpSw.attach(GEAR_UP_SW_SENSE,INPUT_PULLUP);
  gearUpSw.interval(DEBOUNCE_INTERVAL);
  throttleLowSw.attach(THROTTLE_LOW_SW_SENSE,INPUT_PULLUP);
  throttleLowSw.interval(DEBOUNCE_INTERVAL);

  pinMode(AEX_TRIGGER,OUTPUT);
  pinMode(AEX_STATUS_LED,OUTPUT);


  #ifdef LCD_DISPLAY
  #ifdef LCD_DISPLAY_TYPE_ADAFRUIT
  lcd.begin(16, 2); 
  #endif
  #ifdef LCD_DISPLAY_TYPE_DFROBOT
  lcd.init();
  #endif
  
  // Print a message to the LCD.
  initDisplay();

  for (int i = 0; i < STATUS_MSG_Q_LEN; i++)
  {
    gStatusMsgQueue[i].createTime = 0;
  }
  #endif

  lidar.setMsgHdr(gsAltHdr,sizeof(gsAltHdr));
  //initDataStream(&lidar,LIDAR_STREAM_BAUDRATE,"LIDAR");
 
  efis.setMsgHdr(gsSpdHdr,sizeof(gsSpdHdr));
#ifdef SPD_MSG_TRLR
  efis.setMsgTrlr(gsSpdTrlr,sizeof(gsSpdTrlr));
#endif 

  //initDataStream(&efis,EFIS_STREAM_BAUDRATE,"EFIS ");

  /* set up initials status indicators */
  aexStatusLed.turnOff();
  aexStatusLed.setFlash(LED_NOFLASH);  // aex status LED to not blink
  aexStatusLed.setIntensity(AEX_STATUS_LED_INTENSITY);  

  stateCheckTimer.start(STATE_CHECK_MILLIS);
  serialStartCheckTimer.start(SERIAL_CHECK_MILLIS); 
  altTrendTimer.start(ALT_TREND_CHECK_MILLIS);
  
  // set upper and lower limits for the buffered sensors 
  spd.setLimits(SPD_MIN,SPD_MAX);
  alt.setLimits(ALT_MIN,ALT_MAX);

#ifdef AUDIO_ALERTS 
  mp3PlayerSerial.begin(19200);
  mp3PlayerSerial.listen();
  // try five times to init the audio player
  for (int i = 0; i < 5; i++)
  { 
    if (mp3Player.begin(mp3PlayerSerial)) 
    {
      gHaveMp3Player = true; 
      break;
    }
    delay(1000);
  }
  if (!gHaveMp3Player)
  {
    queueFlashMsg(fAudioFailMsg);
  }
  else 
  {
    mp3Player.switchFunction(mp3Player.MUSIC);
    playAudioFile(fAudioEnabled);
  }
  EFIS_STREAM.listen();
#endif
  gAltCheckMillis  = millis();
  #define ALT_CHECK_MILLIS 500   
  gGearCheckMillis  = millis();
  // gear up warning check every 11 secs 
  #define GEAR_CHECK_MILLIS 11000       
  
    
    
}

void loop() // run over and over
{

 /*
  if (throttleLowSw.changed() )
  {
    lcdPrint(0,"sw state change");
    
    if (getSwState(THROTTLE_LOW_SW_SENSE,throttleLowSw.read()) == HIGH)
    {
      lcdPrint(1,"sw state: HIGH");
    }
    else 
    {
      lcdPrint(1,"sw state: LOW  ");
    }

  }

 */ 

  updateInputs();

  updateTimers();
  
  updateLeds();

  updateTrends();

  updateAex(); 
             
  updateDisplay(gCurrentAltitude, gCurrentAirspeed); 

#ifdef AUDIO_ALERTS 
  updateAudio();
#endif  
}

void updateInputs()
{
  int altitude, airspeed ; 
 
  aexEnableSw.update();
  gearUpSw.update();
  throttleLowSw.update();
#if (LIDAR_TYPE == LIDAR_TYPE_SF11_ANALOG) || (LIDAR_TYPE == LIDAR_TYPE_SF30_ANALOG)
  if (!(alt.getEnableState())) { alt.enable();}   // analog alt sensor always enabled 
  altitude = getAltitude();
  if (altitude >= 0)
  {
    alt.update(altitude);
    gCurrentAltitude = alt.getValue(); 
  }
#else  
  lidar.update();
  if (lidar.haveNewMsg())
  {

    
    if (!(alt.getEnableState())) { alt.enable();}  // re-anable the sensor if it was disabled 

    altitude = getAltitude(); 
    if (altitude >= 0)
    {
      alt.update(altitude);
      gCurrentAltitude = alt.getValue();
    }
  }
  else 
  {
    if ((lidar.getMsgMillis() + STREAM_TIMEOUT_MILLIS) < millis())  // stream has timed out
    {
      alt.disable();    // disable the sensor 
      gCurrentAltitude = UNKNOWN;    // alt value is unreliable  
    }
  }
#endif 
  
  efis.update();
  if (efis.haveNewMsg())
  {
    if (!(spd.getEnableState())) 
    { 
      spd.enable(); 
      
    }  // re-enable the sensor if it was disabled 
    airspeed = getAirSpeed();
    if (airspeed >= 0)
    {
      spd.update(airspeed);
      gCurrentAirspeed = spd.getValue();
    }
  }
  else 
  {
    if ((efis.getMsgMillis() + STREAM_TIMEOUT_MILLIS) < millis())  // stream has timed out
    {
      spd.disable();    // disable the sensor 
      gCurrentAirspeed = UNKNOWN;    // spd value is unreliable  
    }
  }
   
  
}

void updateTimers()
{

  statusMsgTimer.update(); 
  stateCheckTimer.update();
  serialStartCheckTimer.update();
  altTrendTimer.update();
}

void updateLeds()
{
  aexStatusLed.update();
}


// update the trends for sensor values we want to keep track of 
void updateTrends()
{
  if (altTrendTimer.isExpired())
  {
   
    if ((gCurrentAltitude != UNKNOWN)  )
    {
      altTrend.update(gCurrentAltitude); 
    }
    altTrendTimer.start(ALT_TREND_CHECK_MILLIS);  // restart the timer 
    if (altTrend.getTrendState() ) 
    {
      switch (altTrend.getTrend())
      {
         case TREND_INCREASING: 
            //queueStatusMsg("Alt increasing");
            break;
         case TREND_DECREASING:
             //queueStatusMsg("Alt decreasing");
            break;           
         default:
             //queueStatusMsg("Alt steady");
             break;    
      }       
    }
  }
}

// check for conditions that would trigger and audio alert and initiate that alert 

void updateAudio()
{
#ifdef AUDIO_ALERTS 
   bool msgPlayed = false; 
   if (millis() > (gAltCheckMillis + ALT_CHECK_MILLIS) )
   {
    gAltCheckMillis = millis();     
    if ((gCurrentAltitude != UNKNOWN) && (gCurrentAltitude < ALT_MAX) && (altTrend.getTrend() == TREND_DECREASING))
    {
        if (gCurrentAltitude <= 100 && !gOneHundredFootPlayed)
       {
         playAudioFile(fOneHundredFoot);
         gOneHundredFootPlayed = true;
         msgPlayed = true; 
       }
        else if (gCurrentAltitude <= 50 && !gFiftyFootPlayed)
        {
         playAudioFile(fFiftyFoot);
         gFiftyFootPlayed = true;
         msgPlayed = true;
        }
        else if (gCurrentAltitude <= 20 && !gTwentyFootPlayed)
        {
          playAudioFile(fTwentyFoot);
          gTwentyFootPlayed = true;
          msgPlayed = true; 
        }
        else if (gCurrentAltitude <= 10 && !gTenFootPlayed)
        {
          playAudioFile(fTenFoot);
          gTenFootPlayed = true;
          msgPlayed = true; 
        } 
        else if (gCurrentAltitude <= 5 && !gFiveFootPlayed)
        {
          playAudioFile(fFiveFoot);
          gFiveFootPlayed = true;        
          msgPlayed = true; 
        }
        else if (gCurrentAltitude <= 3 && !gThreeFootPlayed)
        {
          playAudioFile(fThreeFoot);
          gThreeFootPlayed = true;
          msgPlayed = true; 
        } 
        else if (gCurrentAltitude <= 1 && !gOneFootPlayed)
        {
          playAudioFile(fOneFoot);
          gOneFootPlayed = true;
          msgPlayed = true; 
        }
  
    }


    // reset the msg played switches if climbing, check for gear still down 
    if (gCurrentAltitude != UNKNOWN && (altTrend.getTrend() == TREND_INCREASING) )
    {
       
        if (gCurrentAltitude >= 100)
        {
          gOneHundredFootPlayed = gFiftyFootPlayed = gTwentyFootPlayed = gTenFootPlayed = gFiveFootPlayed = gThreeFootPlayed = gOneFootPlayed = false;
        
          if (  (getSwState(GEAR_UP_SW_SENSE,gearUpSw.read()) != ON) && !gGearDownWarnPlayed)     
          { 
            playAudioFile(fGearDownWarn);
            gGearDownWarnPlayed = true;         
          }
        }  
    }
    
    if (gGearDownWarnPlayed && (getSwState(GEAR_UP_SW_SENSE,gearUpSw.read()) == ON)) gGearDownWarnPlayed = false; 
    
 
   }

   // issue gear up warning every GEAR_CHECK_MILLIS if gear up sw is on, we are descending and below ALT_MAX ft, throttle is retarded, aex is not enabled 
   if (millis() > (gGearCheckMillis + GEAR_CHECK_MILLIS))
   {
     gGearCheckMillis = millis();
     if(gCurrentAltitude != UNKNOWN && 
        gCurrentAltitude < ALT_MAX  && 
        altTrend.getTrend() == TREND_DECREASING &&
        (getSwState(GEAR_UP_SW_SENSE,gearUpSw.read()) == ON) &&
        (!((gAexState == AEX_ARMED) || (gAexState == AEX_ARMED_MIN_SENSOR)) ) &&  //aex not armed  
        (getSwState(THROTTLE_LOW_SW_SENSE,throttleLowSw.read()) == ON) &&
        !msgPlayed  &&    // don't step on other audio messages if played in this cycle         
        1 == 1)
     {
        playAudioFile(fGearUpWarn);   
     }
   }

   // 
    
#endif   
}

// check set aex state based on inputs 
 
void updateAex()
{

  int aexCurrState, aexNewState ;
  aexCurrState = aexNewState = gAexState; 
  
  
  if ( stateCheckTimer.isExpired() )
  {
    //lcdPrint(0,"hndl change");
    switch (aexCurrState)   
    {
       
      case AEX_OFF:
        //lcdPrint(1,"case off");
        if (aexEnableSw.read() == HIGH)
        {
          aexNewState = AEX_ON; 
        }  
        else 
        {
          if (serialStartCheckTimer.isExpired())
          {
            serialStartCheckTimer.start(SERIAL_CHECK_MILLIS);  // restart the timmer
            initAex();  // try to start serial streams (no-op if already started)
          }
        }
        break;
      case AEX_ON:
        //lcdPrint(1,"case on");
        if (aexEnableSw.read()  == LOW )
        {
          aexNewState = AEX_OFF;
        }
        else // aexsw is on, check # active sensors  
        {
            if (getActiveSensors() >= AEX_MIN_SENSOR_CNT) //  ensure we have min# sensors reporting 
            {
              if (gearUpSw.read() == LOW  )  // and gear sw in UP position
              {
                if (getActiveSensors() > AEX_MIN_SENSOR_CNT)
                {
                  aexNewState = AEX_ARMED;   // can move to armed state 
                }
                else 
                {
                  aexNewState = AEX_ARMED_MIN_SENSOR;  // move to armed/warning state 
                }
              }
            }  
        }
        break;
      
      case AEX_ARMED:
        //lcdPrint(1,"case armed");
        if (aexEnableSw.read() == LOW)  // aex sw turned off 
        {
          aexNewState = AEX_OFF;   // move to off state
        }
        else 
        {
          if (gearUpSw.read() == HIGH)  // gear sw not in up position 
          {
            aexNewState = AEX_ON; // disarm 
          }
          else  // aex on, gear sw up 
          {
          
            if (getActiveSensors() <= AEX_MIN_SENSOR_CNT )  // at or below min sensors  
            {
              if (getActiveSensors() < AEX_MIN_SENSOR_CNT)  // below min sensor count 
              {
                aexNewState = AEX_FAIL;  // move to fail state               
              }
              else 
              {
                aexNewState = AEX_ARMED_MIN_SENSOR; 
              }
            }
            else // all systems go, check trigger conditions and trigger gear ext. if met 
            {
              if (checkAexTriggerConditions())
              {
              
                if (gAexTrigger == OFF)
                {
                  setAexTrigger(ON);
                }
              }
            }  
          }
        }
        break;

      // similar to armed state but with warning msg and flashing led 
      case AEX_ARMED_MIN_SENSOR:
        //lcdPrint(1,"case armed min sensor");
        if (aexEnableSw.read() == LOW)  // aex sw turned off 
        {
          aexNewState = AEX_OFF;   // move to off state
        }
        else 
        {
          if (gearUpSw.read() == HIGH)  // gear sw not in up position 
          {
            aexNewState = AEX_ON; // disarm 
          }
          else  // aex on, gear sw up 
          {
          
            if (getActiveSensors() < AEX_MIN_SENSOR_CNT ) //  below min sensor count   
            {
              aexNewState = AEX_FAIL;  // move to fail state               
            }
            if (getActiveSensors() > AEX_MIN_SENSOR_CNT)  //above min sensor count 
            {
                aexNewState = AEX_ARMED ;
            }
            else // all systems go, check trigger conditions and trigger gear ext. if met 
            {
                if (checkAexTriggerConditions())
                {
              
                  if (gAexTrigger == OFF)
                  {
                    setAexTrigger(ON);
                    
                  }
                }
             }
           }  
        }
        
        break;

        
      case AEX_FAIL:
        //lcdPrint(1,"case fail");
        if (aexEnableSw.read() == LOW)
        {
          aexNewState = AEX_OFF; 
        }
        else 
        { 
          if (getActiveSensors() >= AEX_MIN_SENSOR_CNT)
          {
            aexNewState = AEX_ON;  // move from fail state to ON state 
          }
          else    // attempt to get sensors online 
          {
            //Serial.println("trying reinit");
            initAex();
          }
        }
        break;
        
      case AEX_HARD_FAIL:
        if (aexEnableSw.read() == LOW)
        {
          aexNewState = AEX_OFF; 
        }     
        
        // if we get here then something has gone horribly wrong - set unrecoverable fail state (require power-on reset to clear)
        
        default:
          //lcdPrint(0,"badstate");
          aexNewState = AEX_HARD_FAIL;
          break;
    }
    

    stateCheckTimer.start(STATE_CHECK_MILLIS);   // restart the timer for the state check 

    // reset the aex trigger if the gear sw is not in up position 
    if (gAexTrigger == ON && getSwState(GEAR_UP_SW_SENSE,gearUpSw.read()) == OFF )   
    {
      setAexTrigger(OFF);
    }
  }
  
  if (gAexState != aexNewState)  // state has changed due to some event 
  {
    
    //Serial.print("changing states from ") ; Serial.print(gAexState) ; Serial.print(" to "); Serial.println(aexNewState); 
    switch (aexNewState)
    {
      case AEX_OFF:
        handleAexOffEvent();
        break;
      case AEX_ON:
        handleAexOnEvent();
        break;
      case AEX_ARMED:
        handleAexArmEvent();
        break;
      case AEX_ARMED_MIN_SENSOR:
        handleAexArmMinSensorEvent();
        break;
      case AEX_FAIL:
        handleAexFailEvent();
        break;
      default:
        handleAexFailEvent();
        break;
    }

    //lcdPrint(1,"state transition");
    //sprintf(infoMsg,"%0d -> %0d",gAexState,aexNewState);
    //lcdPrint(1,infoMsg);
    gAexState = aexNewState;
    //lcdPrintState();
    //Serial.print("gAexstate:"); Serial.println(gAexState);
  }
 
}

// event handler for transition to ON state  
void handleAexOnEvent()
{
          //Serial.println("in handleAexOnEvent");
          //lcdPrintState("enter on evt");
          setAexTrigger(OFF);
          initAex(); 
          aexStatusLed.turnOn();
          aexStatusLed.setFlash(AEX_STATUS_LED_INIT_FLASH_MILLIS);       

          if (getActiveSensors() < AEX_MIN_SENSOR_CNT)
          {
           queueFlashMsg(fWaitSensorMsg);   // AEX waiting for sensors           
          }
          else
          {
            queueFlashMsg(fEnableMsg);   // AEX Enabled 
#ifdef AUDIO_ALERTS 
            playAudioFile(fAexEnable);
#endif             
          }  
          //lcdPrintState("exit on evt"); 
}

// event handler for transition to OFF state 
void handleAexOffEvent()
{
          setAexTrigger(OFF); 
          aexStatusLed.turnOff(); 
          queueFlashMsg(fDisableMsg);   // AEX Disabled
#ifdef AUDIO_ALERTS 
          playAudioFile(fAexDisable);
#endif 
}

// event handler for transition to FAIL state 
void handleAexFailEvent()
{        

          aexStatusLed.setFlash(AEX_STATUS_LED_FAIL_FLASH_MILLIS);  
          getFlash(gFlashBuf,fFailMsg,sizeof(gFlashBuf));     
          queueStatusMsg(gFlashBuf);  // AEX failed 
}

// event handler for transition to ARMED state 
void handleAexArmEvent()
{
  int sensorCnt = getActiveSensors();  
  setAexArmedLed(sensorCnt);
  queueFlashMsg(fArmMsg);

}



// event handler for transition to ARMED_MIN_SENSOR state 
void handleAexArmMinSensorEvent()
{
  int sensorCnt = getActiveSensors();  
  setAexArmedLed(sensorCnt);
  if (!(spd.getEnableState()) )   // no speed sensor
  {
    queueFlashMsg(fArmNoSpdMsg);    
  }

  else if (!(alt.getEnableState()) )  // no alt sensor 
  {
    queueFlashMsg(fArmNoAltMsg);    
  }
  else 
  {
     queueFlashMsg(fArmWarnMsg);  // default degraded sensor arm msg (shouldn't get here)
  }

}


// set AEX led armed indicators
void setAexArmedLed(int sensorCnt)
{
   if (sensorCnt == AEX_MIN_SENSOR_CNT)
   {
      aexStatusLed.setFlash(AEX_STATUS_LED_WARNING_FLASH_MILLIS);
   }
   else 
   {
      aexStatusLed.setFlash(LED_NOFLASH);
   }
}

// return count of active sensors 
int getActiveSensors()
{
  int count = 0;
  if (alt.isRecent() && alt.getEnableState()) count++; 
  if (spd.isRecent() && spd.getEnableState()) count++ ; 
  
  return count ; 
}

void setAexTrigger(bool triggerState)
{
  if (triggerState == ON)
  {
    gAexTrigger = ON;
    digitalWrite(AEX_TRIGGER,HIGH);
    aexStatusLed.setFlash(AEX_STATUS_LED_DEPLOY_FLASH_MILLIS); 
    queueFlashMsg(fTriggerMsg);  // AEX Triggered
    queueFlashMsg(fExtendMsg);  // Extending Gear...
    #ifdef AUDIO_ALERTS 
    playAudioFile(fGearExtending);
    #endif 
  }
  if (triggerState == OFF)
  {
    if (gAexTrigger == ON)
    { 
          queueFlashMsg(fExtCancelMsg); 
    }
    gAexTrigger = OFF;
    digitalWrite(AEX_TRIGGER,LOW); 
  } 
}

/*  ensure all the conditions are met for the gear to be autoextended:
  - aex is enabled
  - gear sw is UP
  - airspeed is valid
  - altitude is valid
  - airspeed > min airsped threshold and < max airspeed threshold
  - altitude < max max altitude threshole
  - throttle < 1500 rpm sw state is ON
*/
bool checkAexTriggerConditions()
{
 
  bool returnVal = false;
  int altitude,  airspeed;
  altitude = alt.getValue();
  bool altCheck, spdCheck, sensorCheck;
  airspeed = spd.getValue();
  if (alt.isRecent())
  {
   altCheck = altitude >= 0 && 
              altitude < UPPER_AGL_THRESHOLD &&
              altitude >= LOWER_AGL_THRESHOLD ; 
  }
  else 
  {
    altCheck = false;
  }
  
  if (spd.isRecent())
  {
   spdCheck = airspeed > 0 && 
       airspeed > LOWER_AIRSPEED_THRESHOLD &&
       airspeed < UPPER_AIRSPEED_THRESHOLD ;
  }
  else 
  {
    spdCheck = false;
  }
  // if both alt and speed avail, use both 
  if (getActiveSensors() > AEX_MIN_SENSOR_CNT)
  {
    sensorCheck = altCheck && spdCheck;
  }
  else  // otherwise, use either 
  {
    sensorCheck = altCheck || spdCheck;
    
  }


  //Serial.println("Checking aex conditions:");
  //Serial.print("Aex state: "); Serial.println(gAexState);
  //Serial.print("Sensor check:" ); Serial.println(sensorCheck);
  
  //Serial.print("Current airspeed is: "); Serial.println(airspeed);
  //Serial.print("Current alt is: "); Serial.println(altitude);
  //if (gAexState == AEX_OK) {Serial.println("AEX is OK");}
  //if (airspeed > 0) {Serial.println("airspeed > 0");}
  //if (gearUpSw.getState() == ON) {Serial.println("gearupSw is on");}
  //if (altitude > 0 ) {Serial.println("alt > 0");}
  //if (spd.isRecent() ) {Serial.println("airspeed is recent");}
  //if (airspeed > LOWER_AIRSPEED_THRESHOLD ) {Serial.println("speed > lower threshold ");}
  //if (airspeed < UPPER_AIRSPEED_THRESHOLD ) {Serial.println("speed < upper threshold ");}
  //if (altitude < UPPER_AGL_THRESHOLD ) {Serial.println("alt < upper threshold ");}
  //if (throttleLowSw.getState() == ON ) {Serial.println("low throttle sw is ON ");}
  
  if ( ((gAexState == AEX_ARMED) || (gAexState == AEX_ARMED_MIN_SENSOR)) &&
       getSwState(GEAR_UP_SW_SENSE,gearUpSw.read()) == ON &&
       sensorCheck &&
       getSwState(THROTTLE_LOW_SW_SENSE,throttleLowSw.read()) == ON &&
       1 == 1
      )
  {
    //Serial.println("triggering AEX...");
    returnVal = true;
  }

  return returnVal;
}
      
/*
  init aex function in response to aex enable button being pushed

 - ensure we are getting alt and spd data from the two serial streams
 - if either are unavailable set the led to flash quickly
 - if both are good turn led on solid
 
  the two serial streams are assumed to already be active
 
 */
void initAex()
{
      //Serial.println("in iniaex...");
     if (!lidar.status())  // no lidar stream, try to start it 
     {    
//#ifndef SERIAL_MONITOR
       initDataStream(&lidar,LIDAR_STREAM_BAUDRATE,"LIDAR"); 
//#endif
     }
     if (!efis.status())  // no efis stream, try to start it 
     {
       initDataStream(&efis,EFIS_STREAM_BAUDRATE,"EFIS "); 
     }

}

  
// init a serial data stream
void initDataStream(serialWrapper *stream, unsigned long int baudrate,char * streamName)
{
      char msg[LCD_DISPLAY_COLS+1];
      int msgLen = 0;
      //sprintf(msg," %s %d   ",streamName, baudrate);
      //Serial.println(msg);
      //Serial.println("in initdatastream with baudrate"); Serial.println(baudrate);
      Timer timeOutTimer = Timer();
      timeOutTimer.start(STREAM_TIMEOUT_MILLIS);
      stream->begin(baudrate);

      // aex enabled, then start flashing the status led
      //lcdPrint(0,"Waiting connect..");
      while(!stream->status())
      {
        //lcdPrint(0,"wait stream opn");
        timeOutTimer.update();
        if (timeOutTimer.isExpired()) {break;}
        delay(1);
      }
      if (timeOutTimer.isExpired())  // if we timed out, terminate the stream
      {
          //Serial.println("stream init timeout");
          sprintf(msg,"%5s fail init",streamName);
          queueStatusMsg(msg);
          stream->end();
      }
}




    
// read the latest altitude from the lidar
int getAltitude()
{
  int altitude = UNKNOWN;  // default to unknown altitude 
  byte lidarMsg[MAX_MSG_LEN] ;
  int lidarMsgLen;
#if (LIDAR_TYPE == LIDAR_TYPE_SF11_ANALOG) 
  altitude = analogRead(LIDAR_ANALOG_IN) * ALT_FT_CONVERSION_FACTOR;
#else
  lidarMsgLen = lidar.getNewMsg(lidarMsg);
  //Serial.print("in getAltitude, lidarmsglen is: "); Serial.println(lidarMsgLen);
#endif 
  //Serial.println(lidarMsgLen);
  #if LIDAR_TYPE == LIDAR_TYPE_TF03 
  if (lidarMsgLen == ALT_MSG_LEN) //ensure we have the right length of data for an altitude data frame
  {
    // if msg valid get altitude integer from msg and convert from raw sensor units to ft agl
    altitude = int((lidarMsg[ALT_MSB_IDX] * 256 + lidarMsg[ALT_LSB_IDX]) * ALT_FT_CONVERSION_FACTOR); 
    
  }
  #elif LIDAR_TYPE == LIDAR_TYPE_SF11


  if (lidarMsgLen == ALT_MSG_LEN || lidarMsgLen == (ALT_MSG_LEN - 1))
  {

    // no checksum for SF11 messages, do a sanity check on the message content to make sure it 
    // contains ascii numbers, spaces, or decimal points
    int lidarMsgIdx;
    int altCm, altMetre; 
    for (lidarMsgIdx  = ALT_MSB_IDX; lidarMsgIdx < lidarMsgLen; lidarMsgIdx++)
    {
       if (isdigit(lidarMsg[lidarMsgIdx]) || 
           lidarMsg[lidarMsgIdx] == '.' || 
           lidarMsg[lidarMsgIdx] == ' ')
       {
          
        }
       else break;
       if (lidarMsgIdx == (lidarMsgLen - 1 ))  // msg is valid 
       {
          if (sscanf(&lidarMsg[ALT_MSB_IDX],"%3d.%2d",&altMetre,&altCm) )
          {
            altitude = ((altMetre * 100) + altCm) * ALT_FT_CONVERSION_FACTOR;
            // if sf11 returns at or over its max rated range, set altitude to max permitted alt
            altitude = (altitude > SF11_ALT_MAX) ? ALT_MAX: altitude; 
            
          }
       }
    }
     
  }
  #endif
  
  return altitude;
}

  
// read the latest speed from the efis
int getAirSpeed()
{
  int speed = UNKNOWN;
  byte efisMsg[MAX_MSG_LEN] ;
  int efisMsgLen;
  efisMsgLen = efis.getNewMsg(efisMsg);
  //Serial.println("efisMsgLen is:") ; Serial.println(efisMsgLen); 
  if (((efisMsgLen == SPD_MSG_LEN) && (SPD_MSG_LEN > 0))
   || ((SPD_MSG_LEN == 0) && (efisMsgLen > 0))  // if variable-len messages allowed, just ensure message len > 0    
  )  //ensure we have the right length of data for a data frame
  {
    #if EFIS_TYPE == EFIS_TYPE_ASX1
    speed = int((efisMsg[SPD_MSB_IDX] * 256 + efisMsg[SPD_LSB_IDX]) * SPD_KT_CONVERSION_FACTOR);
    #elif (EFIS_TYPE == EFIS_TYPE_SV) || (EFIS_TYPE == EFIS_TYPE_GX3)
    // ensure SV checksum is valid 
    char checkSumBuf[3];
    getCheckSum(checkSumBuf,&efisMsg[SPD_MSG_CHECKSUM_START_IDX],SPD_MSG_CHECKSUM_LEN);
    //Serial.println(checkSumBuf);
    //Serial.println((char *)efisMsg);
    if (strncmp(&efisMsg[SPD_MSG_CHECKSUM_IDX],checkSumBuf,sizeof(checkSumBuf) - 1) == 0)
    {
      // checksum valid, calc speed. sv speed is 4 ascii characters expressed in 1/10'ths of knots
      int spdAccum = 0;
      unsigned int spdMult = 1;
      // note: assumption in line below that MSB_IDX < LSB_IDX 
      for (int i = SPD_LSB_IDX; i >= SPD_MSB_IDX; i--)
      {
        spdAccum = spdAccum + ((efisMsg[i] & 0x000F) * spdMult);   // take bottom nibble and mult by 10**i
        spdMult = spdMult * 10;
      }
      // convert msg speed to knots 
      speed = spdAccum * SPD_KT_CONVERSION_FACTOR;
    }
    #elif EFIS_TYPE == EFIS_TYPE_GRT 
    // handle shadin-type 'Z' variable length message 
    // scan thru message calculating the checksum and looking for the "ZA" (ias) and "ZR" (checksum) sentences 
    int spdLSBidx,spdMSBidx;
    spdLSBidx = spdMSBidx = 0; 
    unsigned int grtCheckSum = 0; 
    byte grtCheckSumBuf[4]; 
    int spdAccum = 0;
    unsigned int spdMult = 1;
    //Serial.print((char *) efisMsg);
    for (int i = 0; i < (efisMsgLen-1); i++)
    {
       grtCheckSum = grtCheckSum + (int)efisMsg[i]; 
       if (efisMsg[i] ==  'Z' && efisMsg[i+1] == 'A')   //ias section
       {
          //Serial.println("found alt section");
          spdMSBidx = i + 2; 
          spdLSBidx = i + 4;
       }
       if (efisMsg[i] == 'Z' && efisMsg[i+1] == 'R')  // checksum section
       {
          // mask off all but lower byte and convert to char decimal 
          grtCheckSum = grtCheckSum & 0x00ff;
          //Serial.print("Checksum: "); Serial.println(grtCheckSum);
          sprintf(grtCheckSumBuf,"%03d",grtCheckSum);
          if (strncmp(&efisMsg[i+2],grtCheckSumBuf,sizeof(grtCheckSumBuf) - 1) == 0)
          {
             // matching checksum, calculate speed in knots 
             if (spdLSBidx > 0)
             {
                for (int j = spdLSBidx; j >= spdMSBidx; j--)
                {
                    spdAccum = spdAccum + ((efisMsg[j] & 0x000F) * spdMult);   // take bottom nibble and mult by 10**i
                    spdMult = spdMult * 10;                 
                }
             }
          }
       }
       
    }
 

    speed = spdAccum; 
    #else
      THROW SYNTAX ERROR IF NO EFIS TYPE MATCHES
    #endif  
    
  }
  return speed;
}



//normalize sw states whether sw is LOW when on or HIGH when on (ie, invert for grounded-when-on type switches)
bool getSwState(int swPin, bool swState)
{
  bool result = swState;
  switch (swPin)
  {
    case AEX_ENABLE_SW:
      result = swState;
      break;
    case GEAR_UP_SW_SENSE:
    case THROTTLE_LOW_SW_SENSE:
      result = !swState;
      break;
    default: 
      result = !swState;
  }
  return result;
  
}

//display welcome message and thresholds
void initDisplay()
{    
      


      #ifdef LCD_DISPLAY
      #ifdef LCD_DISPLAY_TYPE_DFROBOT
      lcd.backlight();
      #endif 
      char  lcdLineBuf[LCD_DISPLAY_COLS + 1];
      lcdClearLine(lcdLineBuf);
      
      lcd.begin(16, 2);
      lcd.clear();

      lcdScroll(F(VERSION_STR));
      delay(1500) ;
  /*    
      lcdScroll("MIN EXT. SPEED:");
      
      lcdClearLine(lcdLineBuf);
      sprintf(lcdLineBuf,"%03d knots",LOWER_AIRSPEED_THRESHOLD);
      lcdPrint(1,lcdLineBuf);
      delay(1500);
      lcdScroll("MAX EXT. SPEED:");
      lcdClearLine(lcdLineBuf);
      sprintf(lcdLineBuf,"%03d knots",UPPER_AIRSPEED_THRESHOLD);
      lcdPrint(1,lcdLineBuf);
      delay(1500);
      lcdScroll("MAX EXT. RPM:");
      lcdClearLine(lcdLineBuf);
      sprintf(lcdLineBuf,"%03d ",UPPER_RPM_THRESHOLD);
      lcdPrint(1,lcdLineBuf);
      delay(1500);
      lcdScroll("MAX EXT. ALT:");
      lcdClearLine(lcdLineBuf);
      sprintf(lcdLineBuf,"%03d ft",UPPER_AGL_THRESHOLD);
      lcdPrint(1,lcdLineBuf);
      delay(1500);
*/
      lcd.clear();
      #endif
}


//
// lcdPrint prints a line of text on the specified line of the display
// note input line is assumed to be a char vector of length LCD_DISPLAY_COLS
//
void lcdPrint(int row, String lineText)
{
      #ifdef LCD_DISPLAY
      lcd.setCursor(0,row);
      lcd.print(lineText);
      return;
      #endif
      
}


void lcdClearLine(char* lineText)
{
      #ifdef LCD_DISPLAY
      for (int i = 0;  i < LCD_DISPLAY_COLS; i++)
      {
        lineText[i] = ' ';
      }
      #endif 
      return ;
     
}


void lcdScroll(String text)
{
      #ifdef LCD_DISPLAY
      lcd.clear();
      lcd.home();
        
      for (int i = 0 ; i < text.length(); i++)
      {

        lcd.print(text[i]);
        if (i >= LCD_DISPLAY_COLS)
        {
          lcd.scrollDisplayLeft();
        };
        delay(250);
        
      }
      #endif
      return;
}

// output the current lidar altitude and skyview airspeed (or error condition if either not available)


void updateDisplay(int altitude , int airspeed)
{
      #ifdef LCD_DISPLAY
      char lcdLineBuf[LCD_DISPLAY_COLS+1];
      char textAlt[5] = "U/A ";
      char textSpd[4] = "U/A"; 
      if (altitude >= 0)
      {
        if (altitude >= ALT_MAX)
        {
          strncpy(textAlt,">MAX",4);  
        }
        else 
        {
          sprintf(textAlt,"%04d",altitude);
        }
      }
      if (airspeed >= 0)
      {
        sprintf(textSpd,"%03d",airspeed);         
      }

      // first line
      lcdClearLine(lcdLineBuf);
      sprintf(lcdLineBuf,"ALT:%04s SPD:%03s",textAlt,textSpd);  
      lcdPrint(LCD_LINE_1,lcdLineBuf);

      // second line  - display oldest undisplayed message if status message display interval has expired 
      statusMsgTimer.update();
      unsigned long oldestTime = ULONG_MAX;
      int displayIdx = -1 ; 
      if ((statusMsgTimer.isRunning() && statusMsgTimer.isExpired()) || !(statusMsgTimer.isRunning() ))
      { 

        for (int i = 0; i < STATUS_MSG_Q_LEN; i++)
        {
          if  (!(gStatusMsgQueue[i].displayed) && (gStatusMsgQueue[i].createTime < oldestTime) && (gStatusMsgQueue[i].createTime > 0))
          {
           oldestTime = gStatusMsgQueue[i].createTime;
           displayIdx = i ;
          }
        } 
        if (displayIdx >= 0)  //
        {
          lcdClearLine(lcdLineBuf);
          lcdPrint(LCD_LINE_2,lcdLineBuf);
          lcdPrint(LCD_LINE_2,gStatusMsgQueue[displayIdx].msg );
          statusMsgTimer.start(STATUS_MSG_DISPLAY_MILLIS);
          gStatusMsgQueue[displayIdx].displayed = true;
         
        }
      } 
      #endif          
}

// add a message to the status message q  
// replace the oldest displayed msg, or if no messages displayed, the oldest message

void queueStatusMsg (char * msg)
{
  #ifdef LCD_DISPLAY
  unsigned long int oldestTime = ULONG_MAX, oldestDisplayedTime = ULONG_MAX;
  int oldestIdx = 0, oldestDisplayedIdx = -1;
  for (int i = 0; i < STATUS_MSG_Q_LEN; i++)
  {
     if (gStatusMsgQueue[i].createTime == 0)
     { 
        oldestDisplayedIdx = i;
        break;
     }
     if (gStatusMsgQueue[i].displayed && gStatusMsgQueue[i].createTime < oldestDisplayedTime) 
     {
        oldestDisplayedTime = gStatusMsgQueue[i].createTime;
        oldestDisplayedIdx = i;
     }
     if (gStatusMsgQueue[i].createTime < oldestTime) 
     {
        oldestTime = gStatusMsgQueue[i].createTime;
        oldestIdx = i;
     }  
  }
  
  if (oldestDisplayedIdx < 0)    // no message displayed, set to oldest 
  {
    oldestDisplayedIdx = oldestIdx;
  }
  // copy new message to oldest displayed msg entry 
  char buf[17];
  //sprintf(buf,"adding at idx %02d",oldestDisplayedIdx);
  //lcdPrint(LCD_LINE_2,buf);
  //delay(1000);
  strncpy(gStatusMsgQueue[oldestDisplayedIdx].msg,msg,LCD_DISPLAY_COLS);
  gStatusMsgQueue[oldestDisplayedIdx].displayed = false;  
  gStatusMsgQueue[oldestDisplayedIdx].createTime = millis();  
  #endif
}
  

//
// print the contents of a TF03 lidar dataframe
//
void printLidarData(char * dataframe)
{
    char strbuf[] = "HHHH CMLO CMHI SSLO SSHI RSV1 RSV3 CHKS \n";
    Serial.println(strbuf);
    sprintf(strbuf, "%02x%02x  %02x  %02x   %02x  %02x  %02x  %02x  %02x",dataframe[0], dataframe[1], dataframe[2],dataframe[3],dataframe[4],dataframe[5],dataframe[6],dataframe[7],dataframe[8]);
    Serial.println(strbuf);
    Serial.println();
    return;
}

// calculate a skyview checksum

void getCheckSum(byte *outBuf, byte *inBuf, int bufLen)
{
  char hexBuf[3];
  unsigned int accum = 0;
  for (int i = 0; i < bufLen; i++)
  {
    accum = accum + (int)inBuf[i];
  }
  accum = accum & 0x00FF;  // keep last byte only
  sprintf(hexBuf,"%02X",accum);
  memcpy(outBuf,hexBuf,sizeof(hexBuf)-1);
}

void getFlash(char * buffer, char * flashStr, int len) 
{
  for (int i = 0; i < len; i++)
  {
    buffer[i] = pgm_read_byte_near(flashStr + i); 
  }
}  

// extract a message from the prom and queue to status line
void queueFlashMsg(char * flashMsg)
{
  #ifdef LCD_DISPLAY
  char buffer[LCD_DISPLAY_COLS]; 
  for (int i = 0; i < LCD_DISPLAY_COLS; i++)
  {
     buffer[i] = pgm_read_byte_near(flashMsg + i); 
  }
  queueStatusMsg(buffer);
  #endif
}


// extract a message from the prom and queue to status line
void playAudioFile(char * flashFileName)
{
  #ifdef AUDIO_ALERTS
  char buffer[AUDIO_FILE_NAME_CNT+1]; 
  for (int i = 0; i < AUDIO_FILE_NAME_CNT; i++)
  {
     buffer[i] = pgm_read_byte_near(flashFileName + i); 
  }
  buffer[AUDIO_FILE_NAME_CNT] = '\0'; 
  mp3PlayerSerial.listen()  ;  // grab the software serial port 
  mp3PlayerSerial.flush()  ;   // turf anything that might be in the queue 
  mp3Player.playSpecFile(buffer);
  EFIS_STREAM.listen();   // restore software serial port 

  #endif
}
